#include <hmg.ch>

#define NTrim( n ) ( LTRIM( STR( n ) ) )

#define Emp2Nil( x ) IF( ISCHAR( x ) .AND. EMPTY( ALLTRIM( x ) ), , x )

#translate ISNIL(  <xVal> ) => ( <xVal> == NIL )
#translate ISARRY( <xVal> ) => ( VALTYPE( <xVal> ) == "A" )
#translate ISCHAR( <xVal> ) => ( VALTYPE( <xVal> ) == "C" )
#translate ISNUMB( <xVal> ) => ( VALTYPE( <xVal> ) == "N" )

#xcommand DEFAULT <v1> TO <x1> [, <vn> TO <xn> ]                        ;
          =>                                                            ;
          IF <v1> == NIL ; <v1> := <x1> ; END                           ;
          [; IF <vn> == NIL ; <vn> := <xn> ; END ]
          


*-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.

FUNCTION HL_StripRZero( ;                 // Clean zero ended string
                        cZEString )   // zero ended string
                        
   LOCAL cRetVal := cZEString

   IF CHR(0) $ cRetVal
      cRetVal := LEFT( cRetVal, AT( CHR( 0 ), cRetVal ) - 1 )
   ENDIF   
   
RETURN cRetVal // HL_StripRZero(

*-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.

FUNCTION HL_Any2Str( ;                       // Convert any type data to String.
                  xAny )

   LOCAL cRVal  := '???',;
         nType  :=  0,;
         aCases := { { "A", { | x | "{...}" } },;                
                     { "B", { | x | "{||}" } },;                
                     { "C", { | x | x }},;
                     { "M", { | x | x   } },;                   
                     { "D", { | x | DTOC( x ) } },;             
                     { "L", { | x | IF( x,".T.",".F.") } },;    
                     { "N", { | x | LTRIM( TRAN( x, IF( INT( x ) # x,;
                                                    "999,999,999,999,999,999.999",;
                                                    "999,999,999,999,999,999,999" ) ) ) } },;
                     { "O", { | x | ":Object:" } },;
                     { "U", { | x | "<NIL>" } } }
                    
   IF (nType := ASCAN( aCases, { | a1 | VALTYPE( xAny ) == a1[ 1 ] } ) ) > 0
      cRVal := EVAL( aCases[ nType, 2 ], xAny )
   ENDIF    
                   
RETURN cRVal // HL_Any2Str()

*-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.

FUNCTION HL_OsVers()

   LOCAL aWinVers := WindowsVersion() 
   
   LOCAL cRetVal := TRIM( aWinVers[ 1 ] ) + ", " + aWinVers[ 3 ]
   
RETURN cRetVal // HL_OsVers()  

*-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.

FUNCTION HL_CallSequence( nBegLevel )

   LOCA nLevel  := nBegLevel,;
        cRetVal := ''

   WHILE !(PROCNAME( nLevel ) == "")
      cRetVal += IF( nLevel > nBegLevel, SPACE( 23 ), '' ) + ;
                 PROCNAME( nLevel ) + ;
                 " (" + HL_Any2Str( PROCLINE( nLevel ) ) + ")" + CRLF
      nLevel++                    
   ENDDO

RETURN cRetVal // HL_CallSequence()

*-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.

/*

   Low-level function GetComputerName() borrowed from MiniGUI-Extended.
   Thanks to community guru Grigory Filatov.
   
*/
      
#pragma BEGINDUMP

#include <windows.h>
#include "hbapi.h"
#include "hbapiitm.h"


// BOOL GetComputerName( LPTSTR lpBuffer, LPDWORD nSize );
// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/sysinfo/base/getcomputername.asp
HB_FUNC( GETCOMPUTERNAME )
{
   TCHAR lpBuffer[ 129 ];
   DWORD nSize = 128;

   GetComputerName( lpBuffer, &nSize );

   hb_retc( lpBuffer );
}

#pragma ENDDUMP      

*-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.

/*

   f.HL_ParsArStr() is a sub-function of f.HL_MsgExt().
   
   Author  : Bicahi Esgici
   
   Purpose : Pars Lines of an array
   
   Syntax  : HL_ParsArStr( <aArray>, <aDelim> ) => <aParsed>
   
   Parameters : <aArray> : Array to parse 
                <aDelim> : Delimiter(s)
                
   Return : <aParsed> : Parsed verison of <aArray>
   
   History :
   
       2008.08 : First Release
   
*/


FUNC HL_ParsArStr(;                                    // Pars Lines of an array
                aArray, aDelim )

   LOCA nDelm  :=  0,;
        cDelm  := '',;
        nLiNo  :=  0,;
        c1Lin  := '',;
        aTemp  := {},;
        a1Arr  := {},;
        nPosNo :=  0,;
        aRVal  := {}

   DEFAULT aDelim TO { CRLF }
   
   IF !ISARRY( aArray )
      aArray := { HL_Any2Str( aArray ) }
   ENDIF   
   
   IF !ISARRY( aDelim )
      aDelim := { HL_Any2Str( aDelim ) }
   ENDIF   

   FOR nDelm := 1 TO LEN( aDelim )
      cDelm := aDelim[ nDelm ]
      FOR nLiNo := 1 TO LEN( aArray )
         c1Lin  := aArray[ nLiNo ]
         a1Arr  := {}
         IF ISCHAR( c1Lin ) .AND. !EMPTY( c1Lin )
            WHILE ( nPosNo := AT( cDelm, c1Lin ) ) > 0
               AADD( a1Arr, LEFT( c1Lin, nPosNo - 1 ) )
               c1Lin := SUBS( c1Lin, nPosNo + LEN( cDelm ) )
            ENDDO
            IF !EMPTY( c1Lin  )
               AADD( a1Arr, c1Lin )
            ENDIF
         ENDIF ISCHAR( c1Lin ) .AND. !EMPTY( c1Lin )
         AEVAL( a1Arr, { | c1 | AADD( aTemp, c1 ) } )
      NEXT nLinNo
      aRVal := ACLONE( aTemp )
      aTemp := {}
   NEXT nDelm

RETU aRVal // HL_ParsArStr()

*-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.

/*

  HMG : Harbour MiniGUI, Win32 GUI library Demo/Sample
 
  Copyright 2002-12 Roberto Lopez <harbourminigui@gmail.com>
 
  HL_MsgExt is Freevare HMG Message function. 
    
  All bug reports and suggestions are welcome.
    
  Developed under Harbour Compiler and 
  MINIGUI - Harbour Win32 GUI library (HMG);
  compiled and linked by MinGW. 
  
  Thanks to "Le Roy" Roberto Lopez.
 
  HL_MsgExt() is a public domain, everyone
  can use any way who like. 
  
  History :
  
     2008.08 : First Release
*/

/*

   f.HL_MsgExt() : Message Extended 
   
   Author  : Bicahi Esgici

   Syntax  : HL_MsgExt( [ <acMsgLines> ],;                          
                          [ <xTitle> ],;
                          [ <acOptions> ],;
                          [ <anBackColor> ],;
                          [ <cFontNameM> ],;
                          [ <nFontSizeM> ],;
                          [ <anFontColorM> ],;
                          [ <cFontNameO> ],;
                          [ <nFontSizeO> ],;
                          [ <xPosition ] )       => <nResult>


   Parameters :

             All parameters are optional.

             <acMsgLines>   : Message Line(s); Char for single, array for multiple lines; default is ""
             <xTitle>       : Tittle for Message box; char; default is ""
             <acOptions>    : Options; Char for single, array for multiple options; default is ""
             <anBackColor>  : Back Color; num array; default is windows default
             <cFontNameM>   : Font Name for message line(s); char; default is windows default
             <nFontSizeM>   : Font Size for message line(s); numeric; default is windows default
             <anFontColorM> : Font Color message line(s);  num array; default is windows default
             <cFontNameO>   : Font Name for option(s); default is windows default
             <nFontSizeO>   : Font Size for option(s); default is windows default
             <aPosition>    : Position relative to the window / form

                              NIL/EMPTY : will be defaulted to : { 0, 0 }

                              Numeric : will be defaulted to : { <this number>, 0 }

                              Array :  may be included 1 to 3 elements :

                                  1 : relativity base: 1: Desktop, 0: parent window/form;  default : 0

                                  2 : If element count is 2, Position code:

                                          0 : Center
                                          1: Up left
                                          2: Up cent
                                          3: Up right
                                          4: Center left
                                          5: Center right
                                          6: Down left
                                          7: Down center
                                          8: Down right

                                          default : 0 ( Center )

                                        else, If element count is 3, this number will be treated as row number
                                              for upper left corner of message box AND

                                  3 Column number for upper left corner of message box

             All parameters may be given as strings (character type). In this case arrays elements may 
             be separated by CRLF or double semicolon (';;') for strings arrays and by comma (',') for
             numeric arrays. Expressions must be specified in their exact form.
             
             
   Return : <nResult> : Number of option selected;
                        When message box closed without selection ( Esc / Alt-F4 ), zero.

   History :
             2008.07 : First Release

*/

FUNC HL_MsgExt( ;                         // Message Extended
               acMsgLines,;  // Message Line(s)
               xTitle,;      // Tittle for Message box
               xOptions,;    // Options
               aBackColor,;  // Back Color
               cFontNameM,;  // Msg. Font Name
               nFontSizeM,;  // Msg. Font Size
               aFontColorM,; // Msg. Font Color
               cFontNameO,;  // Opt. Font Name
               nFontSizeO,;  // Opt. Font Size
               xPosition  )  // Position relative to the window / form

   LOCA aMsgLines := {},;   // Message Lines
        aOptions  := {},;   // Options
        nRVal     :=  0,;
        nLinNo    :=  0,;
        c1Line    := '',;
        nOpts     :=  0,;
        c1Optn    := '',;
        nMxLnLn   :=  0,;  // Max Line Len
        nOptnCo   :=  0,;  // Option Count
        cLblNam   := '',;  // Label Name
        cbtnNam   := '',;  // Button Name
        nButRow   :=  0,;
        nLnLnPx   :=  0    // Message Line Length in pixel

   LOCA nRelBase  :=  0,;
        nMsBxRow  :=  0,;
        nMsBxCol  :=  0,;
        nPosCode  :=  0,;
        cPosCode  := '',;
        nBaseRow  :=  0,;
        nBaseCol  :=  0,;
        nBasHeig  :=  0,;
        nBasWidt  :=  0
        
   LOCA nMLineCo  :=  0,;         // Messages line count
        nMBxHeig  :=  0,;
        nMBxWidt  :=  0,;
        nMxBtnLn  :=  0,;         // Max Button length (in pixel )
        nBtn1Len  :=  0,;
        nBtnTotL  :=  5
        
   LOCA nMFontSiz :=  0,;
        nOFontSiz :=  0,;
        lVertOpts := .F.        

   * * * * * * * * * * * * * * * * * * * * * * * * *
   *
   *  Defaulting parameters      
   *
   * * * * * * * * * * * * * * * * * * * * * * * * *

   acMsgLines  := Emp2Nil( acMsgLines )
   xTitle      := Emp2Nil( xTitle )    
   xOptions    := Emp2Nil( xOptions )   
   aBackColor  := Emp2Nil( aBackColor ) 
   cFontNameM  := Emp2Nil( cFontNameM )
   nFontSizeM  := Emp2Nil( nFontSizeM )
   aFontColorM := Emp2Nil( aFontColorM )
   cFontNameO  := Emp2Nil( cFontNameO )
   nFontSizeO  := Emp2Nil( nFontSizeO )
   xPosition   := Emp2Nil( xPosition ) 

   DEFAULT acMsgLines  TO  {""},;
           xTitle      TO  HL_Any2Str( xTitle ),;
           xOptions    TO  {" Ok "},;              // aBackColor  TO  { 201, 215, 228 },;  // 
           xPosition   TO  { 0, 0 }

   IF ISCHAR( aBackColor )
      aBackColor := HL_ParsArStr( aBackColor , { CRLF, ',' } ) 
      IF LEN( aBackColor ) # 3
        aBackColor := NIL
      ELSE   
        AEVAL( aBackColor, { | c1, i1 | aBackColor[ i1 ] := VAL( c1 ) } )
      ENDIF   
   ENDIF
   
   IF ISCHAR( aFontColorM )
      aFontColorM := HL_ParsArStr( aFontColorM , { CRLF, ',' } ) 
      IF LEN( aFontColorM ) # 3
         aFontColorM := NIL
      ELSE   
         AEVAL( aFontColorM, { | c1, i1 | aFontColorM[ i1 ] := VAL( c1 ) } )
      ENDIF   
   ENDIF
   
   IF ISCHAR( nFontSizeM )
      nFontSizeM := VAL( nFontSizeM )
   ENDIF
      
   IF ISCHAR( nFontSizeO )
      nFontSizeO := VAL( nFontSizeO )
   ENDIF

   IF ISNIL( cFontNameM ) .AND. !ISNIL( nFontSizeM )
      cFontNameM := "Tahoma"
   ENDIF
      
   IF ISNIL( cFontNameO ) .AND. !ISNIL( nFontSizeO )
      cFontNameO := "Tahoma"
   ENDIF
   
   nMFontSiz := MAX( MIN( IF( ISNIL( nFontSizeM ), 9, nFontSizeM ), 32 ), 8 )      
   nOFontSiz := MAX( MIN( IF( ISNIL( nFontSizeO ), 9, nFontSizeO ), 32 ), 8 )
   
   IF !ISNIL( nFontSizeM )     
      nFontSizeM := nMFontSiz 
   ENDIF   
   
   IF !ISNIL( nFontSizeO )     
      nFontSizeO := nOFontSiz 
   ENDIF   
   
   aMsgLines := HL_ParsArStr( acMsgLines, { CRLF, ";;" } )                  // Message lines
   
   nMLineCo := LEN( aMsgLines )

   AEVAL( aMsgLines, { | c1 | nMxLnLn := MAX( nMxLnLn, LEN( c1 ) ) } )   // Max Line Legth of Messages lines

   nLnLnPx :=  nMxLnLn * nMFontSiz * .8                                  // Messages line length by pixel
   
   * * * * * * * * * * * * * * * * * * * * * * * * *
   *
   * Buttons ( Options )
   *
   * * * * * * * * * * * * * * * * * * * * * * * * *
   
   aOptions := HL_ParsArStr( xOptions, { CRLF, ";;" } )                // Options

   nOptnCo :=  LEN( aOptions )                                      // Options (buttons) Count
   
   lVertOpts := ( EMPTY( aMsgLines ) .AND. nOptnCo > 1 )
   
   AEVAL( aOptions,  { | c1 |  nBtn1Len := nOFontSiz * LEN( c1 ) * .8 + 10,;
                               nMxBtnLn := MAX( nMxBtnLn, nBtn1Len ),; 
                               nBtnTotL += nBtn1Len } )
                               

   
   * * * * * * * * * * * * * * * * * * * * * * * * *
   *
   * Form metrics   ( very problematic ! :(
   *
   * * * * * * * * * * * * * * * * * * * * * * * * *
      
   IF lVertOpts 
      nMBxHeig := 40 + ( nOptnCo * 4 * nOFontSiz ) 
      nMBxWidt := MAX( nMxBtnLn, 123 ) + 100 // !!! 123 : Min Window WIDTH !!!
   ELSE   
      nMBxHeig := 40 + ( nMLineCo * 4 * nMFontSiz ) + ( nOFontSiz * 2 )  // * 6
      nMBxWidt := MAX( MAX( nBtnTotL, nLnLnPx ), 123 ) + 10 // !!! 123 : Min Window WIDTH !!!
   ENDIF lVertOpts 
   
   * * * * * * * * * * * * * * * * * * * * * * * * *
   *
   * Position of Message Box
   *
   * * * * * * * * * * * * * * * * * * * * * * * * *

   IF EMPTY( xPosition )
      xPosition := { 0, 0 }
   ENDIF
   
   IF ISCHAR( xPosition )
      xPosition := HL_ParsArStr( xPosition, { CRLF, ',' } ) 
   ENDIF
   
   IF ISARRY( xPosition )
      AEVAL( xPosition, { | x1, i1 | xPosition[ i1 ] := IF( ISCHAR( x1 ), VAL( x1 ), 0 ) } )
   ELSE
      IF ISNUMB( xPosition )
         xPosition := { xPosition, 0 }
      ELSE   
         xPosition := { 0, 0 }
      ENDIF   
   ENDIF

   nRelBase := xPosition[ 1 ]
   
   IF nRelBase < 1     // Parent ( Caller of f.HL_MsgExt() form/window
      nBaseRow :=  ThisWindow.Row
      nBaseCol :=  ThisWindow.Col
      nBasHeig :=  ThisWindow.Height 
      nBasWidt :=  ThisWindow.Width
   ELSE                // Desktop 
      nBasHeig :=  GetDesktopHeight() 
      nBasWidt :=  GetDesktopWidth() 
   ENDIF nRelBase < 1      
   

   IF LEN( xPosition ) > 2
      nMsBxRow := nBaseRow + xPosition[ 2 ]
      nMsBxCol := nBaseCol + xPosition[ 3 ]
   ELSE
   
      nPosCode := xPosition[ 2 ]
      cPosCode := NTrim( nPosCode )
      
      IF cPosCode $ "123"             // Up line
         nMsBxRow :=  nBaseRow + 50
      ELSEIF cPosCode $ "678"         // Down lin
         nMsBxRow :=  nBasHeig - nMBxHeig - 10 + nBaseRow 
      ELSE                            // Center Line
         nMsBxRow :=  ( nBasHeig - nMBxHeig ) / 2 + nBaseRow 
      ENDIF

      IF cPosCode $ "146"             // Left Column
         nMsBxCol :=  nBaseCol + 10
      ELSEIF cPosCode $ "358"         // Rigth Column
         nMsBxCol :=  nBaseCol + nBasWidt / 3 * 2
      ELSE                            // Center Column
         nMsBxCol :=  ( nBasWidt - nMBxWidt ) / 2 + nBaseCol
      ENDIF

   ENDIF LEN( xPosition ) > 2

   DEFINE WINDOW frmHL_MsgExt ;
      AT     nMsBxRow, nMsBxCol ;
      WIDTH  nMBxWidt ;
      HEIGHT nMBxHeig ;
      TITLE  xTitle ;
      MODAL  ; //      NOSIZE ;
      NOSYSMENU ;
      BACKCOLOR aBackColor  // { 201, 215, 228 } { 24, 240, 223 }
      
      ON KEY ESCAPE ACTION frmHL_MsgExt.Release
      
      FOR nLinNo := 1 TO nMLineCo
         cLblNam := 'lbl_' + STRZERO( nLinNo, 2 )
         c1Line  := aMsgLines[ nLinNo ]
         DEFINE LABEL &cLblNam
            ROW       (nLinNo * 3 - 1 ) * nMFontSiz
            COL       0
            VALUE     c1Line
            WIDTH     nMBxWidt
            HEIGHT    nMFontSiz * 2
            FONTNAME  cFontNameM
            FONTSIZE  nFontSizeM
            BACKCOLOR aBackColor
            FONTCOLOR aFontColorM
            CENTERALIGN  .T.
         END LABEL
         
      NEXT nLinNo
   
      IF lVertOpts 
         nButRow  := nOFontSiz * 2
      ELSE
         nButRow := nMBxHeig - 70 - nOFontSiz * 3
      ENDIF lVertOpts 
      
      nBtnSpac := INT( ( nMBxWidt - nBtnTotL ) / ( nOptnCo + 1 ) )
      n1BtnCol := nBtnSpac
      
      FOR nOpts := 1 TO nOptnCo
      
         c1Optn   := aOptions[ nOpts ]
         cbtnNam  := 'btn_'  + STRZERO( nOpts, 2 )
         nBtnWidt := INT( nOFontSiz * ( LEN( c1Optn ) ) * .8 + 10 )         
         
         IF lVertOpts 
            n1BtnCol := ( nMBxWidt - nBtnWidt ) / 2
         ENDIF lVertOpts 
         
         DEFINE BUTTON &cbtnNam
            ROW         nButRow
            COL         n1BtnCol 
            CAPTION     c1Optn
            ACTION      { || nRVal := VAL( RIGHT( this.name, 2 ) ),  frmHL_MsgExt.Release }
            WIDTH       nBtnWidt        
            HEIGHT      nOFontSiz * 2 + 4
            FONTNAME    cFontNameO
            FONTSIZE    nFontSizeO
            CENTERALIGN .T.
         END BUTTON
         
         IF lVertOpts
            nButRow += nOFontSiz * 3
         ELSE          
            n1BtnCol += nBtnWidt + nBtnSpac
         ENDIF lVertOpts 
      
      NEXT nOpts

   END WINDOW // frmHL_MsgExt

   ACTIVATE WINDOW frmHL_MsgExt
   
RETU nRVal // HL_MsgExt()

*-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.

FUNCTION HL_MemVarList(;                     // Build an array with variables recorded as a .mem file
                     cMemFNam )
                     
   LOCAL aRetVal  := {},;
         cMFilStr := '',;
         c1VHeadr := '',;
         c1VName  := '',;
         c1VType  := '',;
         c1VData  := '',;
         x1VValu  := '',;
         n1VLeng  := 0,;
         n1VDeci  := 0,;
         n1VWidt  := 0,;
         nFilPntr := 1,;
         nMFStrLn := 0,;
         nActLen  := 0   // Actual Length ( regarding value )
   
   cMFilStr := MEMOREAD( cMemFNam )
   
   nMFStrLn := LEN ( cMFilStr )
   
   WHILE nFilPntr < nMFStrLn 
    
      c1VHeadr := SUBSTR( cMFilStr, nFilPntr, 20 )
      c1VName  := HL_StripRZero( LEFT( c1VHeadr, 11 ) )
      c1VType  := CHR( ASC(SUBSTR( c1VHeadr, 12, 1 ) ) - 128 )      
      n1VDeci  := ASC( SUBSTR( c1VHeadr, 18, 1 ) )
      
      nFilPntr += 32      
      
      DO CASE 
         CASE c1VType = "L"
              n1VWidt := 1
              c1VData := SUBSTR( cMFilStr, nFilPntr, n1VWidt )
              x1VValu := ASC( c1VData ) > 0
              nActLen := 1
         CASE c1VType = "D"
              n1VWidt := 8
              c1VData := SUBSTR( cMFilStr, nFilPntr, n1VWidt )
              x1VValu := ( CTOF( c1VData ) - 2415021 ) + CTOD( "01/01/1900" ) 
              nActLen := 8
         CASE c1VType = "N"
              n1VWidt := 8
              c1VData := SUBSTR( cMFilStr, nFilPntr, n1VWidt )
              x1VValu := HL_Any2Str( CTOF( c1VData ) )
              nActLen := LEN( x1VValu )              
         OTHERWISE // "C"  
              n1VLeng  := ASC( SUBSTR( c1VHeadr, 17, 1 ) )
              n1VWidt  := n1VDeci * 256 + n1VLeng - 1
              c1VData := SUBSTR( cMFilStr, nFilPntr, n1VWidt )                        
              x1VValu := HL_StripRZero( c1VData )             
              nActLen := LEN( x1VValu )                            
      ENDCASE
      
      AADD( aRetVal, { c1VName,;    // Variable name
                       c1VType,;    // Variable type 
                       nActLen,;    // Actual lenght of variable
                       n1VDeci,;    // Decimal digit count
                       x1VValu } )  // Recorded value of variable
      
      nFilPntr += n1VWidt + IF( c1VType = "C", 1, 0 )
      
   ENDDO   
   
RETURN aRetVal // HL_MemVarList()

*-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.-._.